/*-----------------------------------------------------------------------------
	ClockPod / bZ[W										(c)TORO
-----------------------------------------------------------------------------*/
#define STRICT
#include <windows.h>
#include "CLOCKPOD.H"

const TCHAR *EMonth[] =
	{ T("Jan"), T("Feb"), T("Mar"), T("Apr"), T("May"), T("Jun"),
	  T("Jul"), T("Aug"), T("Sep"), T("Oct"), T("Nov"), T("Dec") };

const TCHAR *EWeek[] = { T("Sun"), T("Mon"), T("Tue"), T("Wed"), T("Thu"), T("Fri"), T("Sat") };
const TCHAR *JWeek[] = { T(""), T(""), T(""), T(""), T(""), T(""), T("y") };

const TCHAR ShortGENGOU[] = T("?MTSHR");
const TCHAR *LongGENGOU[] = {T(""), T(""), T("吳"), T("a"), T(""), T("ߘa")};
const int OffsetGENGOU[] = {0, 1867, 1911, 1925, 1988, 2018};

const TCHAR JapaneseErasPath[] = T("SYSTEM\\CurrentControlSet\\Control\\Nls\\Calendars\\Japanese\\Eras"); // Windows7ȍ~
DWORD ExtGENGOUyear = 0;

#define GENGOU_option_short 0
#define GENGOU_option_long 1
#define GENGOU_max_support 2019

void InitExtGENGOU(void)
{
	HKEY hRegEras;
	int cnt = 0;

	ExtGENGOUyear = MAX32;
	if ( ERROR_SUCCESS != RegOpenKeyEx(HKEY_LOCAL_MACHINE, JapaneseErasPath, 0, KEY_READ, &hRegEras) ){
		return;
	}

	for ( ; ; cnt++ ){			// ݒo ---------------------
		DWORD keySize, valueSize;
		TCHAR keyName[MAX_PATH];	// "1989 01 08"
		TCHAR ErasValue[MAX_PATH];	// "__Heisei_H"
		DWORD Rtyp;
		const TCHAR *ptr;
		DWORD srdata;

		keySize = TSIZEOF(keyName);
		valueSize = TSIZEOF(ErasValue);
		if ( ERROR_SUCCESS != RegEnumValue(hRegEras, cnt, keyName, &keySize, NULL,
				&Rtyp, (BYTE *)ErasValue, &valueSize) ){
			break;
		}
		ptr = keyName;
		srdata = GetNumber(&ptr); // year
		if ( srdata <= GENGOU_max_support ) continue;

		ptr += 1; // '_' skip
		srdata = (srdata * 100) + GetNumber(&ptr); // month
		ptr += 1; // '_' skip
		srdata = (srdata * 100) + GetNumber(&ptr); // day
		if ( ExtGENGOUyear > srdata ) ExtGENGOUyear = srdata;
	}
	RegCloseKey(hRegEras);
}

int GetExtGENGOU(SYSTEMTIME *sTime, TCHAR *dest, TCHAR option)
{
	HKEY hRegEras;
	int cnt = 0, len = 0;
	DWORD srTime;

	if ( ERROR_SUCCESS != RegOpenKeyEx(HKEY_LOCAL_MACHINE, JapaneseErasPath, 0, KEY_READ, &hRegEras) ){
		return 0;
	}
	srTime = (((DWORD)sTime->wYear) * 10000) + (sTime->wMonth * 100) + sTime->wDay;

	for ( ; ; cnt++ ){				// ݒo ---------------------
		DWORD keySize, valueSize;
		TCHAR keyName[MAX_PATH];	// "1989 01 08"
		TCHAR ErasValue[MAX_PATH];	// "__Heisei_H"
		DWORD Rtyp;
		const TCHAR *ptr;
		DWORD srdata;

		keySize = TSIZEOF(keyName);
		valueSize = TSIZEOF(ErasValue);
		if ( ERROR_SUCCESS != RegEnumValue(hRegEras, cnt, keyName, &keySize, NULL,
				&Rtyp, (BYTE *)ErasValue, &valueSize) ){
			break;
		}
		ptr = keyName;
		srdata = GetNumber(&ptr); // year
		if ( srdata <= GENGOU_max_support ) continue;

		ptr += 1; // '_' skip
		srdata = (srdata * 100) + GetNumber(&ptr); // month
		ptr += 1; // '_' skip
		srdata = (srdata * 100) + GetNumber(&ptr); // day
		if ( srTime >= srdata ){
			if ( option == GENGOU_option_short ){
				len = wsprintf(dest, T("%c%02d"),
						ErasValue[(6 / sizeof(TCHAR)) + 2],
						sTime->wYear - srdata / 10000 + 1);
			}else{
				ErasValue[4 / sizeof(TCHAR)] = '\0';
				len = wsprintf(dest, T("%s%02d"),
						ErasValue, sTime->wYear - srdata / 10000 + 1);
			}
		}
	}
	RegCloseKey(hRegEras);
	return len;
}

int GetGENGOU(SYSTEMTIME *sTime, TCHAR *dest, TCHAR option)
{
	DWORD srTime = (((DWORD)sTime->wYear) * 10000) + (sTime->wMonth * 100) + sTime->wDay;

	if ( srTime >= ExtGENGOUyear ){
		if ( ExtGENGOUyear == 0 ){
			InitExtGENGOU();
			if ( srTime < ExtGENGOUyear ) return GetGENGOU(sTime, dest, option);
		}
		return GetExtGENGOU(sTime, dest, option);
	}else if ( srTime < 20190501 ){ // 
		if ( option == GENGOU_option_short ){
			return wsprintf(dest, T("H%02d"), sTime->wYear - 1988);
		}else{
			return wsprintf(dest, T("%02d"), sTime->wYear - 1988);
		}
	}else{ // ߘa
		if ( option == GENGOU_option_short ){
			return wsprintf(dest, T("R%02d"), sTime->wYear - 2018);
		}else{
			return wsprintf(dest, T("ߘa%02d"), sTime->wYear - 2018);
		}
	}
}

int GetUpperGENGOU(const SYSTEMTIME *sTime, TCHAR *dest, const TCHAR option, const TCHAR type)
{
	int GENGOUoffset = ((sizeof(ShortGENGOU) / sizeof(TCHAR)) - 2);

	for (;;){
		if ( (type == ShortGENGOU[GENGOUoffset]) || (GENGOUoffset == 0) ){
			break;
		}
		GENGOUoffset--;
	}

	if ( option == GENGOU_option_short ){
		return wsprintf(dest, T("%c%02d"), ShortGENGOU[GENGOUoffset], sTime->wYear - OffsetGENGOU[GENGOUoffset]);
	} else{
		return wsprintf(dest, T("%s%02d"), LongGENGOU[GENGOUoffset], sTime->wYear - OffsetGENGOU[GENGOUoffset]);
	}
}

//------------------------------------- wpX̃t@C߂
int GetFiles(LPCTSTR *fmt)
{
	TCHAR path[MAX_PATH];
	const TCHAR *p;
	HANDLE hFF;
	WIN32_FIND_DATA findfile;
	int count = 0;

	p = tstrchr(*fmt,'>');
	if ( p == NULL ) return -1;
	memcpy(path,*fmt, TSTROFF(p - *fmt));
	path[p - *fmt] = '\0';
	*fmt = p;
	hFF = FindFirstFile(path, &findfile);
	if (hFF == INVALID_HANDLE_VALUE) return -1;
	do{
		if ( ! (
				(findfile.cFileName[0] == '.') &&
					( (findfile.cFileName[1] == '\0' ) ||
					  ((findfile.cFileName[1] == '.') &&
					   (findfile.cFileName[2] == '\0')) ))  ){
			count++;
		}
	}while(FindNextFile(hFF, &findfile) != FALSE);
	FindClose(hFF);
	return count;
}

//------------------------------------- Kȗ𐶐
int trnd(int valmax)
{
	static int seed = 62628173;

	seed = seed * 5294671 + TimerNowTick + 1;
	return valmax ? (int)((DWORD)(seed >> 8) % (DWORD)valmax) : seed;
}
//------------------------------------- w̓eŃOo͂
void LogWrite(DWORD id, const TCHAR *message)
{
	HANDLE hFile;
	DWORD writesize, size;
	TCHAR buf[1200];

	if ( X_log == 0 ) return;
	tstrcpy(buf, cfgfile);
	tstrcpy(strschr(buf,'.'), T(".LOG"));

	hFile = CreateFile(buf, GENERIC_WRITE, 0, NULL,
			OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if ( hFile == INVALID_HANDLE_VALUE ) return;
	size = 0;
	SetFilePointer(hFile, 0, (LONG *)&size, FILE_END);
	#ifdef UNICODE
	{
		char bufA[2000];

		wsprintf(buf, L"%02d-%02d %02d:%02d\t%x:%s\r\n",
				ltime.wMonth, ltime.wDay, ltime.wHour, ltime.wMinute,
				id, message);
		writesize = UnicodeToUtf8(buf, bufA, sizeof(bufA)) - 1;
		WriteFile(hFile, bufA, writesize, &size, NULL);
	}
	#else
		writesize = wsprintf(buf,"%02d-%02d %02d:%02d\t%x:%s\r\n",
				ltime.wMonth, ltime.wDay, ltime.wHour, ltime.wMinute,
				id, message);
		WriteFile(hFile, buf, writesize, &size, NULL);
	#endif
	CloseHandle(hFile);
	return;
}
//------------------------------------- "%" WJ
TCHAR *tsprintf(TCHAR *dest, const TCHAR *fmt, const TCHAR *params)
{
	TCHAR buf[MESLEN * 2],*destmax;
	TCHAR c;

	destmax = dest + MESLEN - 1;
	for ( ; (c = *fmt) != '\0' ; fmt++ ){
		const TCHAR *paramfmt,*pt;
		INT_PTR paramdata;

		if ( dest >= destmax ){
			dest = destmax;
			break;
		}
		if ( c != '%' ){
			*dest++ = c;
			continue;
		}
		paramfmt = NULL;
		#ifdef _DEBUG
			paramdata = 0;
		#endif
		switch ( *(++fmt) ){
			case '%':
				*dest++ = '%';
				break;
		// N
			case 'y':
				paramfmt = T("%02d");
				paramdata = ltime.wYear % 100;
				break;
			case 'Y':
				paramfmt = T("%04d");
				paramdata = ltime.wYear;
				break;
			case 'g':
				dest += GetGENGOU(&ltime, dest, GENGOU_option_short);
				break;
			case 'G':
				dest += GetGENGOU(&ltime, dest, GENGOU_option_long);
				break;
			case 'j':
				if ( *(fmt + 1) == '\0' ) break;
				dest += GetUpperGENGOU(&ltime, dest, GENGOU_option_short, *(++fmt));
				break;
			case 'J':
				if ( *(fmt + 1) == '\0' ) break;
				dest += GetUpperGENGOU(&ltime, dest, GENGOU_option_long, *(++fmt));
				break;
		// 
			case 'n':
				paramfmt = T("%2d");
				paramdata = ltime.wMonth;
				break;
			case 'N':
				paramfmt = T("%02d");
				paramdata = ltime.wMonth;
				break;
			case 'a':
				paramfmt = EMonth[ltime.wMonth - 1];
				break;
		// 
			case 'd':
				paramfmt = T("%02d");
				paramdata = ltime.wDay;
				break;
		// T
			case 'w':
				paramfmt = EWeek[ltime.wDayOfWeek];
				break;
			case 'W':
				paramfmt = JWeek[ltime.wDayOfWeek];
				break;
		// 
			case 'h':
				paramfmt = T("%2d");
				paramdata = ltime.wHour;
				if ( paramdata >= 12 ) paramdata -= 12;
				break;
			case 'H':
				paramfmt = T("%2d");
				paramdata = ltime.wHour;
				break;
		// 
			case 'm':
				paramfmt = T("%02d");
				paramdata = ltime.wMinute;
				break;
		// b
			case 's':
				paramfmt = T("%02d");
				paramdata = ltime.wSecond;
				break;
		// ʕb
		// 
			case 't':
				paramfmt = (ltime.wHour < 12) ? T("am") : T("pm");
				break;
			case 'T':
				paramfmt = (ltime.wHour < 12) ? T("ߑO") : T("ߌ");
				break;
		// e
			case 'p':
				ReadBatteryInfo();
				switch( PowStatus.ACLineStatus ){
					case 0:
						paramfmt = T("DC");
						break;
					case 1:
						paramfmt = T("AC");
						break;
					default:
						paramfmt = T("?C");
						break;
				}
				break;
			case 'b':
				ReadBatteryInfo();
				if ( PowStatus.BatteryLifePercent > 100 ){
					paramfmt = T("???");
				}else{
					paramfmt = T("%3d");
					paramdata = PowStatus.BatteryLifePercent;
				}
				break;
			case 'B':
				ReadBatteryInfo();
				if ( PowStatus.BatteryFlag != 255 ){
					if ( PowStatus.BatteryFlag & B7 ){
						paramfmt = T("No Battery");
					}else if ( PowStatus.BatteryFlag & B3 ){
						paramfmt = T("Charging");
					}else if ( PowStatus.BatteryFlag & (B1 | B2) ){
						paramfmt = T("Low");
					}
				}
				break;
			case 'r':{
				int i;

				if ( ConnectionList[LastConnection].rasname[0] ){
					paramfmt = ConnectionList[LastConnection].rasname;
					break;
				}

				for ( i = 0 ; i < MAXRAS ; i++ ){
					if ( ConnectionList[i].rasname[0] ){
						paramfmt = ConnectionList[i].rasname;
						break;
					}
				}
				break;
			}
			case 'R':
				fmt++;
				paramfmt = T("%d");
				paramdata = trnd(GetNumber(&fmt));
				fmt--;
				break;

			case 'I':
				paramfmt = IMEstatus;
				break;

			case 'v': {
				TCHAR val[0x100],*valp,*valmax;

				valp = val;
				valmax = val + TSIZEOF(val) - 2;
				for (;;) {
					TCHAR chr;

					fmt++;
					chr = *fmt;
					if ( !Isalnum(chr) ) break;
					if ( valp < valmax ) *valp++ = chr;
				}
				buf[0] = *valp = '\0';
				GetEnvironmentVariable(val, buf, TSIZEOF(buf));
				paramfmt = T("%s");
				paramdata = (INT_PTR)buf;
				fmt--;
				break;
			}

			case 'c':
				paramfmt = T("%d");
				paramdata = GetCPUusing();
				break;
			case 'C':
				paramfmt = T("%03d");
				paramdata = GetCPUusing();
				break;

			case '<':
				fmt++;
				paramfmt = T("%d");
				paramdata = GetFiles(&fmt);
				break;

			case 'A':
				buf[0] = '\0';
				GetWindowText(ahWnd, buf, TSIZEOF(buf));
				paramfmt = T("%s");
				paramdata = (INT_PTR)buf;
				break;

			case 'U':
			// UHn : HDTEMPS.EXE gHDDx\
			// UMx : 
			// UGt : GPUx
			// UGr : GPUgp
			// UP0, UP1 : WinRing0.dll gCPUx\(J@\)
				fmt++;
				if ( *fmt == 'G' ){
					TCHAR type;

					type = *(fmt + 1);
					if ( type != '\0' ){
						fmt++;
					}
					paramfmt = T("%s");
					paramdata = (INT_PTR)buf;
					GetGPUInfomation(type, buf);
				}else if ( *fmt == 'H' ){
					TCHAR drive;

					paramfmt = T("%s");
					paramdata = (INT_PTR)buf;
					buf[0] = '\0';
					drive = *(fmt + 1);
					if ( Isdigit(drive) ) fmt++;
					CallHDTEMP(drive, buf);
				}else if ( *fmt == 'M' ){
					TCHAR type;

					type = *(fmt + 1);
					if ( type != '\0' ){
						fmt += 1;
						paramfmt = T("%s");
						paramdata = (INT_PTR)buf;
						GetMemoryInfo(type, buf);
					}
				}else if ( *fmt == 'P' ){
					fmt++;
					paramfmt = T("%d");
					if ( *fmt == 'F' ){
						GetCpuClock();
						fmt++;
						paramdata = cpuinfo[*fmt - '0'].CurrentMhz;
					}else{
						GetCPUTemp();
						paramdata = (*fmt == '0') ? CPUtemp1 : CPUtemp2;
					}
				}
				break;

			default:
				if ( params == NULL ) break;
				pt = params;
				while ( *pt ){
					const TCHAR *next;
					size_t s;

					next = tstrchr(pt,'\t');
					if ( next ){
						s = next - pt;
						next++;
					}else{
						s = tstrlen(pt);
						next = pt + s;
					}
					if ( *fmt == *pt ){
						s--;
						if ( (dest + s) > destmax ) s = destmax - dest;
						memcpy(dest, pt + 1, TSTROFF(s));
						dest += s;
						break;
					}
					pt = next;
				}
		}
		if ( paramfmt != NULL ) dest += wsprintf(dest, paramfmt, paramdata);
	}
	*dest = '\0';
	*(dest + 1) = '\0';
	return dest;
}
/*-----------------------------------------------------------------------------
	bZ[Wi[
-----------------------------------------------------------------------------*/
void SetMessage(int id, const TCHAR *mes, DWORD flag, const TCHAR *params)
{
	TCHAR buf[MESLEN * 2];
										// ϊ -------------------------
	if ( id == zFlow ) return;	// _Flow ͖[vɂȂ̂Ŗ
	if ( *mes ){
		tsprintf(buf, mes, params);
	}else{ // Message Text
		if ( flag & DMES_TELOP ){ // Telop wL͍̎폜
			ACdeleteMain(id);
			return;
		}
		tstrcpy(buf, S_default);
	}
										// Log o ---------------------------
	if ( flag & DMES_LOG ) LogWrite(id, buf);
										// bZ[W -------------------------
	if ( flag & DMES_SENDCP ) SendPodSlot(T("*"), buf);
	if ( flag & DMES_SENDMSG ) SendMessenger(T("*"), buf);
										// Telop ------------------------------
	if ( flag & DMES_TELOP ){
		if (DispMesStock >= MaxMes){					// X^bNtH
			Z_ExecAction(zFlow, NULL);
		}else{
			int i;

			i = DispMesStock;
			if ( i && (id != zNoID) ){
				if ( flag & DMES_COMPRESS ){	// ނēꊷ
					for ( i = 0 ; i < DispMesStock ; i++ ){
						if (DMessage[i].id == id) break;
					}
				}else if ( flag & DMES_MEARGE ){
					if ( DMessage[i - 1].id == id){
						i--;
					}
				}
			}
			DMessage[i].id = id;
			tstrcpy(DMessage[i].mes, buf);
			tstrcat(DMessage[i].mes, T(" ")); // TelopԂ̋؂
			DMessage[i].flag = flag;
			if ( i >= DispMesStock ){
				DispMesStock++;
				if ( DispMesStock == 1 ){ // TelopJn̑傫gKvH
					if ( (X_telopx < 0) || (X_telopl < 0) ) ResetTime(); //g
				}
			}
		}
	}
										// W\̕ύX ---------------------
	if ( flag & DMES_NEWTITLE ){
		tstrcpy(S_format[textlines - 1], buf);
		ResetTime();
	}
										// Dialog\ -------------------------
	if ( flag & DMES_DIALOG ) MessageBox(hMainWnd, buf, ClockPodStr, MB_OK);
										// Balloon\ ------------------------
	if ( flag & DMES_BALLOON ){
		TrayMessage(hMainWnd, NIM_ADD, Ic_ICON, ClockPodStr, buf);
	}
}
/*-----------------------------------------------------------------------------
	~ςebv폜
-----------------------------------------------------------------------------*/
void DeleteTelop(int no)
{
	if ( DispMesStock == 0 ) return;		// Ȃ
	if ( no >= DispMesStock ) return;	// ݂Ȃ
	ExecDelTelop(no);
	if ( !(--DispMesStock) ){
		ForceDraw = TRUE;
		DispMesX = 0;
		OldaWndLen = 0;
	}
	if ( no < DispMesStock ){
		memmove(&DMessage[no], &DMessage[no+1],
				(DispMesStock - no) * sizeof(DMESSAGE));
	}
}
/*-----------------------------------------------------------------------------
	ebv폜̃ANVs
-----------------------------------------------------------------------------*/
void ExecDelTelop(int no)
{
	TCHAR buf[MESLEN * 2];
	const TCHAR *idname = T("");
	int id;

	id = DMessage[no].id;
	if ((DWORD)id < MESS){
		idname = CustMes[id];
	}else{				// Mail H
		ACOFF offset;
		AC_MAIL *acm;

		for ( offset = AC_mails ; offset ; offset = acm->next ){
			acm = (AC_MAIL *)(cfgac.bottom + offset);
			if ( id == (int)(AC_MAIL *)acm){
				idname = ThTOFFSETSTR(&cfgstr, acm->id);
				break;
			}
		}
	}
	wsprintf(buf, T("1%s\t2%s"), idname, DMessage[no].mes);
	Z_ExecAction(zDelTelop, buf);
}
